#include "hal_rx.h"
#include <stddef.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <util/atomic.h>
#include "ring_buffer.h"


#define SET_GP_OUTPUT(port, p, value) do { \
    if (value) { \
        (port) |= _BV((p)); \
    } else { \
        (port) &= ~_BV((p)); \
    } \
} while (0)


#define RX_DATA_PORT PORTB
#define RX_DATA_DDR DDRB
#define RX_DATA_PIN PINB
#define RX_DATA_P PB5

#define LEDS_PORT PORTD
#define LEDS_DDR DDRD
#define LEDS_P PD6

#define MOTOR_OUTPUT_A_PORT PORTB
#define MOTOR_OUTPUT_A_DDR DDRB
#define MOTOR_OUTPUT_A_P PB0

#define MOTOR_OUTPUT_B_PORT PORTB
#define MOTOR_OUTPUT_B_DDR DDRB
#define MOTOR_OUTPUT_B_P PB1

#define MOTOR_PWM_OUTPUT_PORT PORTB
#define MOTOR_PWM_OUTPUT_DDR DDRB
#define MOTOR_PWM_OUTPUT_P PB2
#define MOTOR_PWM_OCR OCR0A

#define PWM_OUTPUT_PORT PORTD
#define PWM_OUTPUT_DDR DDRD
#define PWM_OUTPUT_P PD5
#define PWM_OCR OCR0B

#define SERVO1_OUTPUT_PORT PORTB
#define SERVO1_OUTPUT_DDR DDRB
#define SERVO1_OUTPUT_P PB3
#define SERVO1_OCR OCR1A

#define SERVO2_OUTPUT_PORT PORTB
#define SERVO2_OUTPUT_DDR DDRB
#define SERVO2_OUTPUT_P PB4
#define SERVO2_OCR OCR1B

#define STATUS_LED_PORT PORTD
#define STATUS_LED_DDR DDRD
#define STATUS_LED_P PD0

#define GPOUT0_PORT PORTD
#define GPOUT0_DDR DDRD
#define GPOUT0_P PD1

#define GPOUT1_PORT PORTA
#define GPOUT1_DDR DDRA
#define GPOUT1_P PA1

#define GPOUT2_PORT PORTA
#define GPOUT2_DDR DDRA
#define GPOUT2_P PA0

#define GPOUT3_PORT PORTD
#define GPOUT3_DDR DDRD
#define GPOUT3_P PD2

#define GPOUT4_PORT PORTD
#define GPOUT4_DDR DDRD
#define GPOUT4_P PD3

#define GPOUT5_PORT PORTD
#define GPOUT5_DDR DDRD
#define GPOUT5_P PD4

#define GPOUT6_PORT PORTD
#define GPOUT6_DDR DDRD
#define GPOUT6_P PD5

// #define GENERAL_PURPOSE_OUTPUT_PORT


static hal_rx_status_t STATUS = HAL_RX_STATUS_OK;
static ring_buffer_t RECEIVE_BUFFER = { 0 };
static hal_rx_callback_t TIMER1_CALLBACK = NULL;
static void *TIMER1_CALLBACK_DATA = NULL;


void hal_rx_init()
{
    // Set clock to 8MHz if not set by fuse bits
    CLKPR = _BV(CLKPCE);
    CLKPR = 0;

    // Enable debug UART, 500000 baud
    // UCSRA = _BV(U2X);
    // UCSRB = _BV(TXEN);
    // UBRRL = 1;

    ring_buffer_init(&RECEIVE_BUFFER);

    // General purpose outputs
    GPOUT0_DDR |= _BV(GPOUT0_P);
    GPOUT1_DDR |= _BV(GPOUT1_P);
    GPOUT2_DDR |= _BV(GPOUT2_P);
    GPOUT3_DDR |= _BV(GPOUT3_P);
    GPOUT4_DDR |= _BV(GPOUT4_P);
    GPOUT5_DDR |= _BV(GPOUT5_P);
    GPOUT6_DDR |= _BV(GPOUT6_P);

    // Set status LED pin as output
    STATUS_LED_DDR |= _BV(STATUS_LED_P);
    STATUS_LED_PORT |= _BV(STATUS_LED_P);

    // Set addressable LEDS pin as output
    LEDS_DDR |= _BV(LEDS_P);

    // Enable PWM outputs A/B for Timer/Counter0, enable fast PWM mode (3)
    TCCR0A = _BV(COM0A1) | _BV(COM0B1) | _BV(WGM01) | _BV(WGM00);
    // Clock source=CLK/1 (PWM frequency 31.250kHz)
    TCCR0B = _BV(CS00);
    // Set output duty cycles to 0
    PWM_OCR = 0;
    MOTOR_PWM_OCR = 0;
    // Enable output drivers
    PWM_OUTPUT_DDR |= _BV(PWM_OUTPUT_P);
    MOTOR_PWM_OUTPUT_DDR |= _BV(MOTOR_PWM_OUTPUT_P);

    // Enable Timer/Counter0 overflow interrupt and USI clocking from timer
    USICR = _BV(USICS0);
    // TIMSK |= _BV(TOIE0);
    TIMSK |= _BV(OCIE0A);

    // Enable outputs for motor
    MOTOR_OUTPUT_A_DDR |= _BV(MOTOR_OUTPUT_A_P);
    MOTOR_OUTPUT_B_DDR |= _BV(MOTOR_OUTPUT_B_P);

    // Enable Timer/Counter1 for servos, mode 14 (Fast PWM, TOP = ICR1)
    TCCR1A = _BV(COM1A1) | _BV(COM1B1) | _BV(WGM11);
    // Clock source=CLK/8, ICR1=20000, PWM frequency 50kHz
    TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS11);
    ICR1 = F_CPU / 8 / 50;
    // Set high=1.5ms
    SERVO1_OCR = 1500;
    SERVO2_OCR = 1500;
    // Enable overflow interrupt
    TIMSK |= _BV(TOIE1);
    // Enable output drivers
    SERVO1_OUTPUT_DDR |= _BV(SERVO1_OUTPUT_P);
    SERVO2_OUTPUT_DDR |= _BV(SERVO2_OUTPUT_P);

    // Enable pull-ups on unused inputs (ISP)
    PORTB |= _BV(PB6) | _BV(PB7);

    sei();
}


void hal_rx_log_write(const char *s)
{
    while (*s)
    {
        while (!(UCSRA & _BV(UDRE)));
        UDR = *s;
        ++s;
    }
}


hal_rx_status_t hal_rx_get_status()
{
    return STATUS;
}


// ISR(TIMER0_OVF_vect)
ISR(TIMER0_COMPA_vect)
{
    volatile const uint8_t data = USIDR;

    if ((USISR & 0x07) == 0)
    {
        if (ring_buffer_is_full(&RECEIVE_BUFFER))
        {
            STATUS = HAL_RX_STATUS_BUFFER_OVERFLOW;
            return;
        }

        ring_buffer_push(&RECEIVE_BUFFER, data);
    }
}


// Every 20ms
ISR(TIMER1_OVF_vect)
{
    static uint8_t counter = 0;

    ++counter;

    if (counter == 50)
    {
        counter = 0;
    }

    if (TIMER1_CALLBACK != NULL)
    {
        TIMER1_CALLBACK(TIMER1_CALLBACK_DATA, counter * 20);
    }
}


bool hal_rx_has_rf_data()
{
    bool result;

    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        result = !ring_buffer_is_empty(&RECEIVE_BUFFER);
    }

    return result;
}


uint8_t hal_rx_receive_rf_data()
{
    uint8_t result;

    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        result = ring_buffer_pop(&RECEIVE_BUFFER);
    }

    return result;
}


void hal_rx_set_status_led(bool is_enabled)
{
    if (is_enabled)
    {
        STATUS_LED_PORT &= ~_BV(STATUS_LED_P);
    }
    else
    {
        STATUS_LED_PORT |= _BV(STATUS_LED_P);
    }
}


void hal_rx_set_timer_callback(hal_rx_callback_t callback, void *data)
{
    TIMER1_CALLBACK = callback;
    TIMER1_CALLBACK_DATA = data;
}


void hal_rx_drive_motor(int8_t speed)
{
    if (speed > 0)
    {
        MOTOR_OUTPUT_A_PORT |= _BV(MOTOR_OUTPUT_A_P);
        MOTOR_OUTPUT_B_PORT &= ~_BV(MOTOR_OUTPUT_B_P);
        MOTOR_PWM_OCR = speed * 2;
    }
    else if (speed < 0)
    {
        MOTOR_OUTPUT_A_PORT &= ~_BV(MOTOR_OUTPUT_A_P);
        MOTOR_OUTPUT_B_PORT |= _BV(MOTOR_OUTPUT_B_P);
        MOTOR_PWM_OCR = -speed * 2 - 1;
    }
    else
    {
        MOTOR_PWM_OCR = 0;
    }
}


void hal_rx_drive_servo1(int16_t t_us)
{
    SERVO1_OCR = t_us;
}


void hal_rx_drive_servo2(int16_t t_us)
{
    SERVO2_OCR = t_us;
}


void hal_rx_set_pwm_output(uint8_t duty_cycle)
{
    PWM_OCR = duty_cycle;
}


void hal_rx_set_gp_output(uint8_t index, bool value)
{
    switch (index)
    {
        case 0:
            SET_GP_OUTPUT(GPOUT0_PORT, GPOUT0_P, value);
            break;

        case 1:
            SET_GP_OUTPUT(GPOUT1_PORT, GPOUT1_P, value);
            break;

        case 2:
            SET_GP_OUTPUT(GPOUT2_PORT, GPOUT2_P, value);
            break;

        case 3:
            SET_GP_OUTPUT(GPOUT3_PORT, GPOUT3_P, value);
            break;

        case 4:
            SET_GP_OUTPUT(GPOUT4_PORT, GPOUT4_P, value);
            break;

        case 5:
            SET_GP_OUTPUT(GPOUT5_PORT, GPOUT5_P, value);
            break;

        case 6:
            SET_GP_OUTPUT(GPOUT6_PORT, GPOUT6_P, value);
            break;
    }
}